package com.loafle.overflow.central.proxy;

import com.google.gson.internal.Primitives;
import com.google.protobuf.ByteString;
import com.loafle.overflow.core.exception.OverflowException;

import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.type.JavaType;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.lang.reflect.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by crusader on 17. 6. 30.
 */
@Component
public class ServiceInvoker {
  @Autowired
  private ApplicationContext context;

  private Map<String, Cache> serviceCacheMap;

  @Autowired
  private ObjectMapper objectMapper;

  public ServiceInvoker() {
    serviceCacheMap = new HashMap<>();
  }

  private static class Cache {
    private ServiceCache serviceCache;
    private Map<String, MethodCache> methodCacheMap;

    private Cache(Object bean) {
      serviceCache = new ServiceCache();
      serviceCache.bean = bean;

      methodCacheMap = new HashMap<>(10);
    }

    private Object getBean() {
      return serviceCache.bean;
    }

    private MethodCache getMethodCache(String methodName) throws NoSuchMethodException {
      MethodCache methodCache = methodCacheMap.get(methodName);
      if (null != methodCache) {
        return methodCache;
      }

      Method method = getMethod(methodName);
      if (null == method) {
        throw new NoSuchMethodException();
      }
      methodCache = new MethodCache();
      methodCacheMap.put(methodName, methodCache);

      methodCache.method = method;
      methodCache.returnClazz = method.getReturnType();

      int paramCount = method.getParameterCount();
      if (0 == paramCount) {
        return methodCache;
      }

      methodCache.parameterTypes = method.getGenericParameterTypes();

      return methodCache;
    }

    private Method getMethod(String methodName) {
      // Class<?> clazz = AopUtils.getTargetClass(serviceCache.bean);
      Class<?> clazz = serviceCache.bean.getClass();
      Method[] methods = clazz.getMethods();

      Method targetMethod = null;
      for (Method m : methods) {
        if (methodName.equals(m.getName())) {
          targetMethod = m;
          break;
        }
      }
      return targetMethod;
    }

    private static class ServiceCache {
      private Object bean;
      private Map<String, MethodCache> methodCacheMap;
    }

    private static class MethodCache {
      private Method method;
      private Class returnClazz;
      private Type[] parameterTypes;
    }
  }

  private Cache getServiceCache(String serviceName) throws java.lang.NoSuchMethodException {
    Cache serviceCache = serviceCacheMap.get(serviceName);
    if (null != serviceCache) {
      return serviceCache;
    }
    try {
      Object bean = context.getBean(serviceName);
      serviceCache = new Cache(bean);
      serviceCacheMap.put(serviceName, serviceCache);
    } catch (BeansException e) {
      throw new NoSuchMethodException();
    }

    return serviceCache;
  }

  private Object getValue(Type parameterType, String json) throws IllegalArgumentException {
    JavaType targetType = objectMapper.getTypeFactory().constructType(parameterType);
    if (!Primitives.isPrimitive(parameterType) && !parameterType.getTypeName().equals(String.class.getName())) {
      try {
        return objectMapper.readValue(json, targetType);
      } catch (IOException e) {
        throw new IllegalArgumentException();
      }
    }
    return objectMapper.convertValue(json, targetType);
  }

  private Object[] getParameters(Type[] parameterTypes, List<ByteString> params) throws IllegalArgumentException {
    if (null == parameterTypes || null == params) {
      return null;
    }

    if (parameterTypes.length != params.size()) {
      throw new IllegalArgumentException();
    }

    Object[] result = new Object[parameterTypes.length];

    for (int i = 0; i < parameterTypes.length; i++) {
      result[i] = getValue(parameterTypes[i], params.get(i).toStringUtf8());
    }

    return result;
  }

  public String invoke(String serviceName, String methodName, List<ByteString> params)
      throws NoSuchMethodException, IllegalAccessException, OverflowException {
    Cache serviceCache = getServiceCache(serviceName);
    Cache.MethodCache methodCache = serviceCache.getMethodCache(methodName);

    if (null == methodCache.parameterTypes) {
      if (null != params && 0 < params.size()) {
        throw new IllegalArgumentException(
            String.format("size of parameters is not valid target[%d], source[%d]", 0, params.size()));
      }
    } else {
      if (methodCache.parameterTypes.length != params.size()) {
        throw new IllegalArgumentException(String.format("size of parameters is not valid target[%d], source[%d]",
            methodCache.parameterTypes.length, params.size()));
      }
    }

    Object[] parameters = null;
    if (null != methodCache.parameterTypes && 0 < methodCache.parameterTypes.length) {
      parameters = getParameters(methodCache.parameterTypes, params);
    }

    Object result = null;
    String jsonInString = null;
    try {
      result = methodCache.method.invoke(serviceCache.getBean(), parameters);
    } catch (IllegalAccessException e) {
      throw new IllegalAccessException(String.format("method[%s.%s] is not valid", serviceName, methodName));
    } catch (InvocationTargetException e) {
      Throwable t = e.getTargetException();
      if (t instanceof OverflowException) {
        throw (OverflowException) t;
      }
      throw new OverflowException("internal error", e);
    }

    try {
      jsonInString = objectMapper.writeValueAsString(result);
    } catch (IOException e) {
      throw new OverflowException("InternalError", e);
    }

    return jsonInString;
  }

}
