package com.loafle.overflow.central.redis.service;

import com.google.gson.Gson;
import com.loafle.overflow.central.commons.service.MessagePublisher;
import com.loafle.overflow.core.exception.OverflowException;
import com.loafle.overflow.core.model.PublishMessage;
import com.loafle.overflow.model.member.Member;
import com.loafle.overflow.service.central.member.MemberService;

import org.codehaus.jackson.map.ObjectMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.listener.ChannelTopic;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

@Service
public class RedisMessagePublisher implements MessagePublisher {
  private static final String CHANNEL_WEBAPP = "/webapp";
  private static final String CHANNEL_NOAUTH_PROBE = "/auth";
  private static final String CHANNEL_PROBE = "/probe";

  @Autowired
  private RedisTemplate<String, Object> redisTemplate;
  @Autowired
  private ObjectMapper objectMapper;
  @Autowired
  private MemberService memberService;

  private Map<String, ChannelTopic> topics;

  public RedisMessagePublisher() {
  }

  public RedisMessagePublisher(final RedisTemplate<String, Object> redisTemplate,
      final Map<String, ChannelTopic> topics) {
    this.redisTemplate = redisTemplate;
    this.topics = topics;
  }

  public void publishToDomainMembers(final Long domainID, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.MEMBER);
    message.setTargets(getMemberListByDomainID(domainID));
    this.publish(CHANNEL_WEBAPP, message, method, params);
  }

  public void publishToDomainMembersByProbeKey(final String probeKey, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.MEMBER);
    message.setTargets(getMemberListByProbeKey(probeKey));
    this.publish(CHANNEL_WEBAPP, message, method, params);
  }

  public void publishToMember(final String memberID, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.MEMBER);
    message.addTarget(memberID);
    this.publish(CHANNEL_WEBAPP, message, method, params);
  }

  public void publishToMemberSession(final String memberSessionID, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.MEMBER_SESSION);
    message.addTarget(memberSessionID);
    this.publish(CHANNEL_WEBAPP, message, method, params);
  }

  public void publishToNoAuthProbe(final String tempProbeKey, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.PROBE);
    message.addTarget(tempProbeKey);
    this.publish(CHANNEL_NOAUTH_PROBE, message, method, params);
  }

  public void publishToProbe(final String probeKey, final String method, final Object... params)
      throws OverflowException {
    PublishMessage message = new PublishMessage();
    message.setTargetType(PublishMessage.TargetType.PROBE);
    message.addTarget(probeKey);
    this.publish(CHANNEL_PROBE, message, method, params);
  }

  @Cacheable("memberListByDomain")
  protected List<String> getMemberListByDomainID(final Long domainID) throws OverflowException {
    return this.getMemberList(memberService.readAllByDomainID(domainID));
  }

  @Cacheable("memberListByProbeKey")
  protected List<String> getMemberListByProbeKey(final String probeKey) throws OverflowException {
    return this.getMemberList(memberService.readAllByProbeKey(probeKey));
  }

  protected List<String> getMemberList(final List<Member> members) {
    List<String> results = new ArrayList<>(members.size());
    for (Member member : members) {
      results.add(member.getEmail());
    }

    return results;
  }

  protected void publish(final String channel, PublishMessage message, final String method, Object... params) {
    ChannelTopic topic = this.topics.get(channel);
    message.setMessage(new PublishMessage.PublishMessageBody(method, this.getMessageBody(params)));

    try {
      String json = this.objectMapper.writeValueAsString(message);
      redisTemplate.convertAndSend(topic.getTopic(), json);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  protected List<String> getMessageBody(final Object... params) {
    List<String> results = new ArrayList<>(params.length);
    for (Object param : params) {
      if (param.getClass().equals(String.class)) {
        results.add((String) param);
      } else {
        try {
          String json = this.objectMapper.writeValueAsString(param);
          results.add(json);
        } catch (IOException e) {
          e.printStackTrace();
        }    
      }
    }
    return results;
  }

}
